# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Page and top-level layout widget.
"""

import os
from os import path
from wehjit.compat import json
from wehjit import base, builtins, staticdir
from wehjit.state import *

class Page(base.Page):
    pass

builtins.register(Page)


class PageApp(base.Page):

    style_vars = (
        ('color', '#000'),
        ('color_bar', '#fff'),

        ('colorbg', '#fff'),
        ('colorbg_code', '#eee'),
        ('colorbg_bar', '#446'),
        ('colorbg_hover_menu', '#669'),
        ('colorbg_target_menu', '#966'),

        ('colorborder_menu', '#000'),

        ('font_family', 'sans-serif'),
        ('font_size', '11pt'),
        ('font_size_mono', '10pt'),

        ('height_bar', '2em'),

        ('padding_menu', '0 0.5em'),
        ('padding_input', '2px'),
        ('width_content', '50em'),
    )

    style_global = (
        ('body', (
            ('font-size', '%(font_size)s'),
            ('font-family', '%(font_family)s'),
            ('color', '%(color)s'),
            ('background-color', '%(colorbg)s'),
        )),

        ('input, select, option, textarea, button', (
            ('font-family', 'monospace'),
            ('font-size', '%(font_size_mono)s'),
            ('line-height', '%(height_bar)s'),
        )),

        ('code', (
            ('background-color', '%(colorbg_code)s'),
            ('padding', '1px'),
        )),

        ('code, pre', (
            ('font-family', 'monospace'),
            ('font-size', '%(font_size_mono)s'),
        )),

        ('label.associate', (
            ('-moz-appearance', 'button'),
            ('padding', '0.1em 0.2em'),
            ('display', 'inline-block'),
        )),

    )

    javascript = """
    var Wehjit = {
        bases: {},

        plugins: {},

        on_load: function() {
            console.log('Wehjit.on_load()');
            wehjit_init.each(Wehjit.each_plugin);
        },

        each_plugin: function(plugin) {
            Wehjit.plugins[plugin.klass] = plugin;
            plugin.instances = [];
            plugin.mooclass = Wehjit.bases[plugin.uses];
            var selector = plugin.tag + '.' + plugin.klass;
            $$(selector).each(function(el) {
                Wehjit.each_instance(plugin, el);
            });
        },

        each_instance: function(plugin, el) {
            console.log('Wehjit.each_instance(%s, %s)', plugin, el);
            plugin.instances.push(
                new plugin.mooclass(plugin.tag, plugin.klass, el)
            );
        },

    };

    window.addEvent('load', Wehjit.on_load);

    Wehjit.bases.Widget = new Class({
        initialize: function(tag, klass, el) {
            console.log('Widget.initialize(%s, %s, %s)', tag, klass, el);
            this.tag = tag;
            this.klass= klass;
            this.el = $(el);
            this.id = this.el.id;
            console.log('    tag = %s', this.tag);
            console.log('    klass = %s', this.klass);
            console.log('    id = %s', this.id);
            console.assert($type(this.el) == 'element');
            console.assert(this.tag == this.el.tagName);
            console.assert(this.el.hasClass(this.klass));
            if (this.add_all_events) {
                this.add_all_events();
            }
            if (this.post_init) {
                this.post_init();
            }

        },

        add_event: function(event) {
            console.log('Widget.add_event(%s)', event);
            var callback = this['on_' + event];
            if (callback && this.target) {
                console.log('    Adding %s event on %s', event, this.id);
                console.log('    target = %s', this.target);
                this.target.addEvent(event, callback.bind(this));
            }
        },
    });


    Wehjit.bases.Form = new Class({
        Extends: Wehjit.bases.Widget,

        add_all_events: function() {
            console.log('Form.add_all_events(), id=%s', this.id);
            this.target = this.el;
            ['submit', 'reset', 'change'].each(this.add_event, this);
        },

        on_submit: function() {
            alert('submit');
            return false;
        },
    });


    Wehjit.bases.FormField = new Class({
        Extends: Wehjit.bases.Widget,

        add_all_events: function() {
            console.log('FormField.add_all_events(), id=%s', this.id);
            var fields = this.el.getElements('input, button, select, textarea');
            if (fields.length == 1) {
                this.field = fields[0];
                this.target = this.field;
            }
            else if (fields.length > 1) {
                this.fields = fields;
                this.target = this.fields;
            }
            ['change', 'focus', 'blur'].each(this.add_event, this);
        },
    });
    """

    @StaticProp
    def inline_js(self):
        return 'var wehjit_init = %s;' % json.dumps(
            list(self._iter_js_classes())
        )

    require_js = ('mootools-core-nc.js',)

    @classmethod
    def register_assets(cls, assets):
        for name in os.listdir(staticdir):
            f = path.join(staticdir, name)
            if path.isfile(f):
                assets.register(f)

    def __init__(self, **kw):
        super(PageApp, self).__init__(**kw)
        self.form = self.new('Form', id='form')
        self.topbar = self.new('TopBar', id='topbar')
        self.menuset = self.new('MenuSet', id='menuset')
        self.menu = self.new('Menu', id='menu')
        self.actions = self.new('Actions', id='actions')
        self.view = self.new('View', id='view')
        self.h1 = self.new('H1', id='h1')
        self.add(self.form)
        self.form.add(self.topbar, self.view)
        self.topbar.add(self.menuset, self.actions)
        self.menuset.add(self.menu)
        self.view.add(self.h1)


builtins.register(PageApp)


class H1(base.Widget):
    """
    An <h1> element with the page title.
    """

    label = Dynamic('label')

    @DynamicProp
    def page_title(self):
        if self.page is None:
            return
        return self.page.title

    xml = """
    <h1
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
        py:content="label or page_title"
    />
    """

    style = (
        ('', (
            ('font-size', '1.5em'),
            ('margin', '0.25em 0'),
        )),
    )

builtins.register(H1)


class TopBar(base.Container):
    """
    A fixed-position div/bar at top of page.
    """

    base_id_from = None

    style = (
        ('', (
            ('position', 'fixed'),
            ('height', '%(height_bar)s'),
            ('top', '0'),
            ('left', '0'),
            ('right', '0'),
            ('z-index', '50'),
            ('line-height', '%(height_bar)s'),
            ('color', '%(color_bar)s'),
            ('background', '%(colorbg_bar)s'),
        )),
    )

builtins.register(TopBar)


class Actions(base.Container):
    base_id_from = None

    style = (
        ('', (
            ('text-align', 'right'),
            ('height', '100%%'),
        )),
        ('> label', (
            ('margin', '0 0.5em'),
        )),
        ('> button', (
            ('margin', '0 0.5em'),
        )),
    )

builtins.register(Actions)


class View(base.Container):
    """
    Scrollable viewport beneath a `TopBar`.
    """

    base_id_from = None

    style = (
        ('', (
            ('padding', '0 1em'),
            ('overflow', 'auto'),
            ('position', 'fixed'),
            ('top', '%(height_bar)s'),
            ('right', '0'),
            ('bottom', '0'),
            ('left', '0'),
        )),
    )

builtins.register(View)



# MenuSet, Menu, and MenuItem share some common style:
ul_style = (
    ('list-style-type', 'none'),
    ('margin', '0'),
    ('padding', '0'),
)

li_style = (
    ('padding', '0'),
    ('margin', '0'),
    ('cursor', 'default'),
    ('color', '%(color_bar)s'),
    ('background', '%(colorbg_bar)s'),
)

li_cell_style = (
    ('padding', '%(padding_menu)s'),
    ('line-height', '%(height_bar)s'),
    ('color', '%(color_bar)s'),
    ('white-space', 'nowrap'),
    ('text-decoration', 'none'),
    ('display', 'block'),
)


class MenuSet(base.Container):
    """
    A group of menus, usually first child in a `TopBar`.
    """

    base_id_from = None

    xml = """
    <ul
        xmlns:py="http://genshi.edgewall.org/"
        id="${id}"
        class="${css_classes}"
    >
    <li py:for="child in children" py:if="child" py:replace="child.generate()" />
    </ul>
    """

    style = (
        ('', ul_style + (
            ('float', 'left'),
        )),

        ('> li', (
            ('float', 'left'),
        )),
    )

builtins.register(MenuSet)


class Menu(base.Container):
    """
    A menu within a `MenuSet`.
    """

    label = Dynamic('label', default='Menu Label')
    title = Dynamic('title')

    xml = """
    <li
        xmlns:py="http://genshi.edgewall.org/"
        id="${id}"
        class="${css_classes}"
        title="${title}"
    >
    <span py:content="label" />
    <ul>
    <li py:for="child in children" py:replace="child.generate()" />
    </ul>
    </li>
    """

    style = (
        ('', li_style),

        (':hover', (
            ('background-color', '%(colorbg_hover_menu)s'),
        )),

        ('> span', li_cell_style),

        ('> ul', ul_style + (
            ('position', 'fixed'),
            ('display', 'none'),
            ('z-index', '60'),
            ('border-width', '1px'),
            ('border-style', 'solid'),
            ('border-color', '%(colorborder_menu)s'),
            ('background-color', '%(colorbg_bar)s'),
        )),

        (':hover > ul', (
            ('display', 'block'),
        )),
    )

builtins.register(Menu)


class MenuItem(base.Widget):
    """
    A menu-item within a `Menu`.
    """

    label = Dynamic('label', default='MenuItem Label')
    href = Dynamic('href')
    title = Dynamic('title')

    @DynamicProp
    def target(self):
        return self.relurl(self.href)

    @StaticProp
    def page_url(self):
        if self.page is None:
            return
        return self.page.url

    xml = """
    <li
        xmlns:py="http://genshi.edgewall.org/"
        id="${id}"
        class="${css_classes}"
        title="${title}"
    >
    <a
        href="${target}"
        class="${(None, 'target')[href == page_url]}"
        py:content="label"
    />
    </li>
    """

    style = (
        ('', li_style),

        ('a', li_cell_style),

        ('a.target', (
            ('background-color', '%(colorbg_target_menu)s'),
        )),

        (':hover > a', (
            ('background', '%(colorbg_hover_menu)s'),
        )),
    )

builtins.register(MenuItem)


class QuickJump(base.Widget):
    """
    A select-dropdown to move between anchors in a page.
    """

    label = Dynamic('label', default='Jump:')
    options = Dynamic('options', default=tuple())

    xml = """
    <label
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
    >
    ${label}
    <select name="${name}">
    <option
        py:for="(optvalue, optlabel) in options"
        value="#${optvalue}"
        py:content="optlabel"
    />
    </select>
    </label>
    """

    js_class = 'QuickJump'
    javascript = """
    Wehjit.bases.QuickJump = new Class({
        Extends: Wehjit.bases.FormField,

        post_init: function() {
            this.start();
        },

        start: function() {
            if (this.timerid) {
                this.stop();
                console.assert(! this.timerid);
            }
            this.timerid = this.poll.periodical(250, this);
        },

        stop: function() {
            this.timerid = $clear(this.timerid);
        },

        poll: function() {
            if (this.field.value != window.location.hash) {
                this.field.value = window.location.hash;
            }
        },

        on_change: function() {
            window.location.hash = this.field.value;
        },

        on_focus: function() {
            this.stop();
        },

        on_blur: function() {
            this.start();
        },

    });
    """

builtins.register(QuickJump)
